{*
 * << P o w e r P d f >> -- PdfDoc.pas
 *
 * Copyright (c) 1999-2001 Takezou. <takeshi_kanno@est.hi-ho.ne.jp>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Library General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or any
 * later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Library general Public License for more
 * details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library.
 *
 * Create 2000.09.10
 *
 *}
unit PdfDoc;

interface

//{DEFINE NOIMAGE}

// if use "FlateDecode" compression, comment out the next line.
// (this unit and PdfTypes.pas)
{$DEFINE NOZLIB}

uses
  SysUtils, Classes, PdfTypes, Dialogs
  {$IFNDEF NOIMAGE}
  , Graphics
  {$ENDIF}
  ;

type
  {*
   * The pagemode determines how the document should appear when opened.
   *}
  TPdfPageMode = (pmUseNone,
                  pmUseOutlines,
                  pmUseThumbs,
                  pmFullScreen);

  {*
   * The line cap style specifies the shape to be used at the ends of open
   * subpaths when they are stroked.
   *}
  TLineCapStyle = (lcButt_End,
                   lcRound_End,
                   lcProjectingSquareEnd);

  {*
   * The line join style specifies the shape to be used at the corners of paths
   * that are stroked.
   *}
  TLineJoinStyle = (ljMiterJoin,
                    ljRoundJoin,
                    ljBevelJoin);

  {*
   * The text rendering mode determines whether text is stroked, filled, or used
   * as a clipping path.
   *}
  TTextRenderingMode = (trFill,
                        trStroke,
                        trFillThenStroke,
                        trInvisible,
                        trFillClipping,
                        trStrokeClipping,
                        trFillStrokeClipping,
                        trClipping);

  TPdfHeader = class(TObject)
  protected
    procedure WriteToStream(const AStream: TStream);
  end;

  TPdfTrailer = class(TObject)
  private
    FAttributes: TPdfDictionary;
    FXrefAddress: integer;
  protected
    procedure WriteToStream(const AStream: TStream);
  public
    constructor Create(AObjectMgr: TPdfObjectMgr);
    destructor Destroy; override;
    property XrefAddress: integer read FXrefAddress write FXrefAddress;
    property Attributes: TPdfDictionary read FAttributes;
  end;

  TPdfXrefEntry = class(TObject)
  private
    FEntryType: string;
    FByteOffset: integer;
    FGenerationNumber: integer;
    FValue: TPdfObject;
    function GetAsString: string;
  public
    constructor Create(AValue: TPdfObject);
    destructor Destroy; override;
    property EntryType: string read FEntryType write FEntryType;
    property ByteOffset: integer read FByteOffSet write FByteOffset;
    property GenerationNumber: integer
                          read FGenerationNumber write FGenerationNumber;
    property AsString: string read GetAsString;
    property Value: TPdfObject read FValue;
  end;

  TPdfXref = class(TPdfObjectMgr)
  private
    FXrefEntries: TList;
    function GetItem(ObjectID: integer): TPdfXrefEntry;
    function GetItemCount: integer;
  protected
    procedure WriteToStream(const AStream: TStream);
  public
    constructor Create;
    destructor Destroy; override;
    procedure AddObject(AObject: TPdfObject); override;
    function GetObject(ObjectID: integer): TPdfObject; override;
    property Items[ObjectID: integer]: TPdfXrefEntry read GetItem;
    property ItemCount: integer read GetItemCount;
  end;

  TPdfCanvas = class;
  TXObjectID = integer;
  TPdfInfo = class;
  TPdfCatalog = class;
  TPdfFont = class;

  {$IFDEF NOZLIB}
  TPdfCompressionMethod = (cmNone);
  {$ELSE}
  TPdfCompressionMethod = (cmNone, cmFlateDecode);
  {$ENDIF}

  TPdfDoc = class(TComponent)
  private
    FRoot: TPdfCatalog;
    FCurrentPages: TPdfDictionary;
    FCanvas: TPdfCanvas;
    FHeader: TPdfHeader;
    FTrailer: TPdfTrailer;
    FXref: TPdfXref;
    FInfo: TPdfInfo;
    FHasDoc: boolean;
    FFontList: TList;
    FXObjectList: TPdfArray;
    FDefaultPageWidth: Word;
    FDefaultPageHeight: Word;
    FCompressionMethod: TPdfCompressionMethod;
    function GetCanvas: TPdfCanvas;
    function GetInfo: TPdfInfo;
    function GetRoot: TPdfCatalog;
  protected
    procedure CreateInfo;
    function CreateCatalog: TPdfDictionary;
    function CreateFont(FontName: string): TPdfFont;
    function CreatePages(Parent: TPdfDictionary): TPdfDictionary;
  public
    procedure RegisterXObject(AObject: TPdfXObject; AName: string);
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure NewDoc;
    procedure FreeDoc;
    procedure AddPage;
    {$IFNDEF NOIMAGE}
    procedure AddImage(AName: string; AImage: TGraphic; ImageClassName: string);
    {$ENDIF}
    procedure SaveToStream(AStream: TStream);
    function GetFont(FontName: string): TPdfFont;
    function GetXObject(AName: string): TPdfXObject;
    function AddOutlineEntry(AParent: TPdfOutlineEntry; ATitle: string;
      AXPos, AYPos: Single; AOpened: boolean): TPdfOutlineEntry;
    property HasDoc: boolean read FHasDoc;
    property Canvas: TPdfCanvas read GetCanvas;
    property Info: TPdfInfo read GetInfo;
    property Root: TPdfCatalog read GetRoot;
  published
    property DefaultPageWidth: word read FDefaultPageWidth write FDefaultPageWidth;
    property DefaultPageHeight: word read FDefaultPageHeight write FDefaultPageHeight;
    property CompressionMethod: TPdfCompressionMethod
       read FCompressionMethod write FCompressionMethod;
  end;

  TPdfCanvasAttribute = class(TObject)
  private
    FWordSpace: Single;
    FCharSpace: Single;
    FFontSize: Single;
    FFont: TPdfFont;
    FHorizontalScaling: Word;
    procedure SetWordSpace(Value: Single);
    procedure SetCharSpace(Value: Single);
    procedure SetFontSize(Value: Single);
    procedure SetHorizontalScaling(Value: Word);
    procedure SetFont(Value: TPdfFont);
  public
    function TextWidth(Text: string): Single;
    function MesureText(Text: string; Width: Single): integer;
    property WordSpace: Single read FWordSpace write SetWordSpace;
    property CharSpace: Single read FCharSpace write SetCharSpace;
    property HorizontalScaling: Word read FHorizontalScaling
      write SetHorizontalScaling;
    property FontSize: Single read FFontSize write SetFontSize;
    property Font: TPdfFont read FFont write SetFont;
  end;

  TPdfCanvas = class(TObject)
  private
    FContents: TPdfStream;
    FPage: TPdfDictionary;
    FPdfDoc: TPdfDoc;
    FAttr: TPdfCanvasAttribute;
    procedure SetPageWidth(AValue: integer);
    procedure SetPageHeight(AValue: integer);
    function GetPageWidth: Integer;
    function GetPageHeight: Integer;
    function GetColorStr(Color: TColor): string;
  protected
  public
    constructor Create(APdfDoc: TPdfDoc);
    destructor Destroy; override;

    {* Special Graphics State *}
    procedure GSave;                                             {  q   }
    procedure GRestore;                                          {  Q   }
    procedure Concat(a, b, c, d, e, f: Single);                  {  cm  }

    {* General Graphics State *}
    procedure SetFlat(flatness: Byte);                           {  i   }
    procedure SetLineCap(linecap: TLineCapStyle);                {  J   }
    procedure SetDash(aarray: array of Byte; phase: Byte);       {  d   }
    procedure SetLineJoin(linejoin: TLineJoinStyle);             {  j   }
    procedure SetLineWidth(linewidth: Single);                   {  w   }
    procedure SetMiterLimit(miterlimit: Byte);                   {  M   }

    {* Paths *}
    procedure MoveTo(x, y: Word);                                {  m   }
    procedure LineTo(x, y: Word);                                {  l   }
    procedure CurveToC(x1, y1, x2, y2, x3, y3: Word);            {  c   }
    procedure CurveToV(x2, y2, x3, y3: Word);                    {  v   }
    procedure CurveToY(x1, y1, x3, y3: Word);                    {  y   }
    procedure Rectangle(x, y, width, height: Word);              {  re  }
    procedure Closepath;                                         {  h   }
    procedure NewPath;                                           {  n   }
    procedure Stroke;                                            {  S   }
    procedure ClosePathStroke;                                   {  s   }
    procedure Fill;                                              {  f   }
    procedure Eofill;                                            {  f*  }
    procedure FillStroke;                                        {  B   }
    procedure ClosepathFillStroke;                               {  b   }
    procedure EofillStroke;                                      {  B*  }
    procedure ClosepathEofillStroke;                             {  b*  }
    procedure Clip;                                              {  W   }
    procedure Eoclip;                                            {  W*  }

    {* Test state *}
    procedure SetCharSpace(charSpace: Single);                   {  Tc  }
    procedure SetWordSpace(wordSpace: Single);                   {  Tw  }
    procedure SetHorizontalScaling(hScaling: Word);              {  Tz  }
    procedure SetLeading(leading: Single);                       {  TL  }
    procedure SetFontAndSize(fontname: string; size: Single);    {  Tf  }
    procedure SetTextRenderingMode(mode: TTextRenderingMode);    {  Tr  }
    procedure SetTextRise(rise: Word);                           {  Ts  }
    procedure BeginText;                                         {  BT  }
    procedure EndText;                                           {  ET  }
    procedure MoveTextPoint(tx, ty: Word);                       {  Td  }
    procedure SetTextMatrix(a, b, c, d, x, y: Word);             {  Tm  }
    procedure MoveToNextLine;                                    {  T*  }
    procedure ShowText(s: string);                               {  Tj  }
    procedure ShowTextNextLine(s: string);                       {  '   }

    {* external objects *}
    procedure ExecuteXObject(xObject: string);                   {  Do  }

    {* Device-dependent color space operators *}
    procedure SetRGBFillColor(Value: TColor);                    {  rg  }
    procedure SetRGBStrokeColor(Value: TColor);                  {  RG  }

    procedure SetPage(APage: TPdfDictionary);
    procedure SetFont(AName: string; ASize: Single);
    procedure DrawXObject(X, Y, AWidth, AHeight: Single;
       AXObjectName: string);
    function TextWidth(Text: string): Single;
    function MesureText(Text: string; AWidth: Single): integer;
    function GetDoc: TPdfDoc;

    property Attribute: TPdfCanvasAttribute read FAttr;
    property Contents: TPdfStream read FContents;
    property Page: TPdfDictionary read FPage;
    property PageWidth: integer read GetPageWidth write SetPageWidth;
    property PageHeight: integer read GetPageHeight write SetPageHeight;
  end;

  TPdfDictionaryWrapper = class(TPersistent)
  private
    FData: TPdfDictionary;
    function GetHasData: boolean;
  protected
    procedure SetData(AData: TPdfDictionary); virtual;
    property Data: TPdfDictionary read FData write SetData;
    property HasData: boolean read GetHasData;
  end;

  TPdfInfo = class(TPdfDictionaryWrapper)
  private
    function GetAuthor: string;
    procedure SetAuthor(Value: string);
    function GetCreationDate: TDateTime;
    procedure SetCreationDate(Value: TDateTime);
    function GetCreator: string;
    procedure SetCreator(Value: string);
    function GetKeywords: string;
    procedure SetKeywords(Value: string);
    function GetSubject: string;
    procedure SetSubject(Value: string);
    function GetTitle: string;
    procedure SetTitle(Value: string);
    function GetModDate: TDateTime;
    procedure SetModDate(Value: TDateTime);
  public
    property Author: string read GetAuthor write SetAuthor;
    property CreationDate: TDateTime read GetCreationDate write SetCreationDate;
    property Creator: string read GetCreator write SetCreator;
    property Keywords: string read GetKeywords write SetKeywords;
    property ModDate: TDateTime read GetModDate write SetModDate;
    property Subject: string read GetSubject write SetSubject;
    property Title: string read GetTitle write SetTitle;
  end;

  TPdfCatalog = class(TPdfDictionaryWrapper)
  private
    FOutlines: TPdfDictionary;
    FPages: TPdfDictionary;
    procedure SetPageMode(Value: TPdfPageMode);
    procedure SetUseOutlines(Value: boolean);
    procedure SetPages(APage: TPdfDictionary);
    function GetPageMode: TPdfPageMode;
    function GetUseOutlines: boolean;
    function GetOutlines: TPdfDictionary;
    function GetPages: TPdfDictionary;
  protected
    procedure CreateOutlines;
  public
    property UseOutlines: boolean read GetUseOutlines write SetUseOutlines;
    property PageMode: TPdfPageMode read GetPageMode write SetPageMode;
    property Outlines: TPdfDictionary read GetOutlines;
    property Pages: TPdfDictionary read GetPages write SetPages;
  end;

  TPDF_STR_TBL = record
    KEY: string;
    VAL: string;
  end;
  TPDF_INT_TBL = record
    KEY: string;
    VAL: integer;
  end;

  TPdfFont = class(TPdfDictionaryWrapper)
  private
    FName: string;
  protected
    procedure AddStrElements(ADic: TPdfDictionary; ATable: array of TPDF_STR_TBL);
    procedure AddIntElements(ADic: TPdfDictionary; ATable: array of TPDF_INT_TBL);
  public
    constructor Create(AXref: TPdfXref; AName: string); virtual;
    function GetCharWidth(AText: string; APos: integer): integer; virtual;
    property Name: string read FName;
  end;

  {$IFNDEF NOIMAGE}
  TPdfImageCreator = class(TPersistent)
  public
    function CreateImage(AXref: TPdfXref; AImage: TGraphic): TPdfXObject; virtual;
    constructor Create;
  end;
  {$ENDIF}

implementation

const
  PDF_PAGE_MODE_NAMES: array[0..3] of string = ('UseNone',
                                              'UseOutlines',
                                              'UseThumbs',
                                              'FullScreen');

{ Utility functions }

procedure _Pages_AddKids(AParent: TPdfDictionary; AKid: TPdfDictionary);
var
  FKids: TPdfArray;
begin
  // adding page object to the parent pages object.
  FKids := AParent.PdfArrayByName('Kids');
  FKids.AddItem(AKid);
  AParent.PdfNumberByName('Count').Value := FKids.ItemCount;
end;

function _Page_GetResources(APage: TPdfDictionary; AName: string): TPdfDictionary;
var
  FResources: TPdfDictionary;
begin
  FResources := APage.PdfDictionaryByName('Resources');
  result := FResources.PdfDictionaryByName(AName);
end;

procedure _OutlineEntry_AddChild(AParent: TPdfDictionary;
  AEntry: TPdfOutlineEntry; AOpened: integer);
var
  OldLastEntry: TPdfOutlineEntry;

  procedure IncCount(AEntry: TPdfDictionary);
  var
    TmpEntry: TPdfDictionary;
    FCountObj: TPdfNumber;
    FOpenObj: TPdfNumber;
    FChildCount: integer;
  begin
    // calculate new chiledren count. if zero (and has children means outlines
    // is closed) then do nothing.
    TmpEntry := AEntry;
    while TmpEntry <> nil do
    begin
      with TmpEntry do
      begin
        // get count attribute from parent entry, if null then create.
        FCountObj := PdfNumberByName('Count');
        if FCountObj = nil then
        begin
          FCountObj := TPdfNumber.CreateNumber(0);
          AddItem('Count', FCountObj);
        end;

        // get open (internal) attribute from parent entry, if null then create.
        FOpenObj := PdfNumberByName('Opened');
        if FOpenObj = nil then
        begin
          if FCountObj.Value > 0 then
            FOpenObj := TPdfNumber.CreateNumber(PDF_ENTRY_OPENED)
          else
            FOpenObj := TPdfNumber.CreateNumber(PDF_ENTRY_CLOSED);
          AddInternalItem('Opened', FOpenObj);
        end;
        FChildCount := Abs(FCountObj.Value);
        inc(FChildCount);
        if FOpenObj.Value = PDF_ENTRY_OPENED then
          FCountObj.Value := FChildCount
        else
          FCountObj.Value := -FChildCount;
      end;
      TmpEntry := TmpEntry.PdfDictionaryByName('Parent');
    end;
  end;
begin
  OldLastEntry := TPdfOutlineEntry(AParent.ValueByName('Last'));
  if OldLastEntry <> nil then
  begin
    AEntry.AddItem('Prev', OldLastEntry);
    OldLastEntry.AddItem('Next', AEntry);
  end
  else
    AParent.AddItem('First', AEntry);
  AParent.AddItem('Last', AEntry);

  IncCount(AParent);

  AEntry.AddItem('Parent', AParent);
  AEntry.AddInternalItem('Opened', TPdfNumber.CreateNumber(AOpened));
end;

function _OutlineEntry_Create(AParent: TPdfDictionary; ATitle: string;
  ADest: TPdfDictionary; AXPos, AYPos: Single; AOpened: integer): TPdfOutlineEntry;
var
  FEntry: TPdfOutlineEntry;
  FDest: TPdfArray;
begin
  FEntry := TPdfOutlineEntry.CreateDictionary(AParent.ObjectMgr);
  FEntry.ObjectMgr.AddObject(FEntry);
  With FEntry do
  begin
    AddItem('Title', TPdfText.CreateText(ATitle));
    FDest := TPdfArray.CreateArray(FEntry.ObjectMgr);
    with FDest do
    begin
      AddItem(ADest);
      AddItem(TPdfName.CreateName('XYZ'));
      AddItem(TPdfReal.CreateReal(AXPos));
      AddItem(TPdfReal.CreateReal(AYPos));
      AddItem(TPdfNumber.CreateNumber(0));
    end;
    AddItem('Dest', FDest);
    _OutlineEntry_AddChild(AParent, FEntry, AOpened);
  end;
  result := FEntry;
end;

function FloatToStrR(Value: Extended): string;
begin
  result := FloatToStr(Trunc(Value * 100 + 0.5) / 100);
end;

{ TPdfHeader }

procedure TPdfHeader.WriteToStream(const AStream: TStream);
begin
  _WriteString('%PDF-1.2 '#13#10, AStream);
end;

{ TPdfTrailer }

procedure TPdfTrailer.WriteToStream(const AStream: TStream);
begin
  _WriteString('trailer' + CRLF, AStream);
  FAttributes.WriteToStream(AStream);
  _WriteString(CRLF + 'startxref' + CRLF, AStream);
  _WriteString(IntToStr(FXrefAddress) + CRLF, AStream);
  _WriteString('%%EOF' + CRLF, AStream);
end;

constructor TPdfTrailer.Create(AObjectMgr: TPdfObjectMgr);
begin
  inherited Create;
  FAttributes := TPdfDictionary.CreateDictionary(AObjectMgr);
  FAttributes.AddItem('Size', TPdfNumber.CreateNumber(0));
end;

destructor TPdfTrailer.Destroy;
begin
  FAttributes.Free;
end;

{ TPdfXrefEntry }

constructor TPdfXrefEntry.Create(AValue: TPdfObject);
begin
  FByteOffset := -1;
  if AValue <> nil then
  begin
    FEntryType := PDF_IN_USE_ENTRY;
    FGenerationNumber := AValue.GenerationNumber;
    FValue := AValue;
  end
  else
  begin
    FEntryType := PDF_FREE_ENTRY;
    FGenerationNumber := 0;
  end;
end;

destructor TPdfXrefEntry.Destroy;
begin
  if FEntryType = PDF_IN_USE_ENTRY then
    FValue.Free;
end;

function TPdfXrefEntry.GetAsString: string;
  function FormatIntToString(Value: integer; Len: integer): string;
  var
    S: string;
    i, j: integer;
  begin
    result := '';
    if Value < 0 then
      S := '0'
    else
      S := IntToStr(Value);
    i := Len - Length(S);
    for j := 0 to i - 1 do
      result := result + '0';
    result := result + S;
  end;
begin
  result := FormatIntToString(FByteOffset, 10) +
            ' ' +
            FormatIntToString(FGenerationNumber, 5) +
            ' ' +
            FEntryType;
end;

{ TPdfXref }

constructor TPdfXref.Create;
var
  RootEntry: TPdfXrefEntry;
begin
  FXrefEntries := TList.Create;
  RootEntry := TPdfXrefEntry.Create(nil);
  RootEntry.GenerationNumber := PDF_MAX_GENERATION_NUM;
  FXrefEntries.Add(RootEntry);
end;

destructor TPdfXref.Destroy;
var
  i: integer;
begin
  for i := 1 to FXrefEntries.Count - 1 do
    GetItem(i).Free;
  FXrefEntries.Free;
  inherited;
end;

procedure TPdfXref.AddObject(AObject: TPdfObject);
var
  ObjectNumber: integer;
  XrefEntry: TPdfXrefEntry;
begin
  // register object to xref table, and set objectnumber.
  if AObject.ObjectType <> otDirectObject then
    raise EPdfInvalidOperation.Create('');
  XrefEntry := TPdfXrefEntry.Create(AObject);
  ObjectNumber := FXrefEntries.Add(XrefEntry);
  AObject.SetObjectNumber(ObjectNumber);
end;

function TPdfXref.GetItem(ObjectID: integer): TPdfXrefEntry;
begin
  result := TPdfXrefEntry(FXrefEntries.Items[ObjectID]);
end;

function TPdfXref.GetItemCount: integer;
begin
  Result := FXrefEntries.Count;
end;

function TPdfXref.GetObject(ObjectID: integer): TPdfObject;
begin
  result := GetItem(ObjectID).Value;
end;

procedure TPdfXref.WriteToStream(const AStream: TStream);
var
  i: integer;
  S: string;
  Count: integer;
begin
  Count := FXrefEntries.Count;
  S := 'xref' +
       CRLF +
       '0 ' +
       IntToStr(Count) +
       CRLF;
  for i := 0 to Count - 1 do
    S := S + Items[i].AsString + CRLF;
  _WriteString(S, AStream);
end;

{ TPdfDoc }
constructor TPdfDoc.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FHasDoc := false;
  FCanvas := TPdfCanvas.Create(Self);
  FDefaultPageWidth := PDF_DEFAULT_PAGE_WIDTH;
  FDefaultPageHeight := PDF_DEFAULT_PAGE_HEIGHT;
  FInfo := nil;
  FRoot := nil;
end;

function TPdfDoc.GetCanvas: TPdfCanvas;
begin
  if not HasDoc then
    raise EPdfInvalidOperation.Create('Document is null');
  result := FCanvas;
end;

function TPdfDoc.GetInfo: TPdfInfo;
begin
  if not HasDoc then
    raise EPdfInvalidOperation.Create('this method can not use this state..');
  if FInfo = nil then
    CreateInfo;
  result := FInfo;
end;

function TPdfDoc.GetRoot: TPdfCatalog;
begin
  if not HasDoc then
    raise EPdfInvalidOperation.Create('this method can not use this state..');
  result := FRoot;
end;

destructor TPdfDoc.Destroy;
begin
  FreeDoc;
  FCanvas.Free;
  inherited;
end;

function TPdfDoc.CreateCatalog: TPdfDictionary;
begin
  // create catalog object and register to xref.
  result := TPdfDictionary.CreateDictionary(FXref);
  FXref.AddObject(result);
  result.AddItem('Type', TPdfName.CreateName('Catalog'));
  FTrailer.Attributes.AddItem('Root', result);
end;

function TPdfDoc.CreateFont(FontName: string): TPdfFont;
var
  PdfFont: TPdfFont;
begin
  // create new font (not regist to xref -- because font object registed by
  // TPdfFont).
  PdfFont := TPdfFont(FindClass(FontName).Create);
  if PdfFont = nil then
    raise Exception.Create('InvalidFontName:' + FontName);
  result := PdfFont.Create(FXref, FontName);
  result.Data.AddItem('Name',
    TPdfName.CreateName('F' + IntToStr(FFontList.Count)));
  FFontList.Add(Result);
end;

procedure TPdfDoc.RegisterXObject(AObject: TPdfXObject; AName: string);
begin
   // check object and register it.
   if AObject = nil then
     raise EPdfInvalidValue.Create('Error AObject is null');
   if _GetTypeOf(AObject.Attributes) <> 'XObject' then
     raise EPdfInvalidValue.Create('Error AObject is null');
   if AObject.ObjectType <> otIndirectObject then
     FXref.AddObject(AObject);
   if AObject.Attributes.ValueByName('Name') = nil then
   begin
     if GetXObject(AName) <> nil then
       raise EPdfInvalidValue.Createfmt('dupulicate name: %s', [AName]);
     FXObjectList.AddItem(AObject);
     AObject.Attributes.AddItem('Name', TPdfName.CreateName(AName));
   end;
end;

procedure TPdfDoc.CreateInfo;
var
  FInfoDictionary: TPdfDictionary;
begin
  FInfoDictionary := TPdfDictionary.CreateDictionary(FXref);
  FXref.AddObject(FInfoDictionary);
  FInfoDictionary.AddItem('Producer', TPdfText.CreateText(POWER_PDF_VERSION_TEXT));
  FTrailer.Attributes.AddItem('Info', FInfoDictionary);
  FInfo := TPdfInfo.Create;
  FInfo.SetData(FInfoDictionary);
end;

function TPdfDoc.CreatePages(Parent: TPdfDictionary): TPdfDictionary;
begin
  // create pages object and register to xref.
  result := TPdfDictionary.CreateDictionary(FXref);
  FXref.AddObject(result);
  with result do
  begin
    AddItem('Type', TPdfName.CreateName('Pages'));
    AddItem('Kids', TPdfArray.CreateArray(FXref));
    AddItem('Count', TPdfNumber.CreateNumber(0));
  end;

  if (Parent <> nil) and (_GetTypeOf(Parent) = 'Pages') then
    _Pages_AddKids(Parent, result)
  else
    FRoot.Pages := result;
end;

function TPdfDoc.GetFont(FontName: string): TPdfFont;
var
  FFont: TPdfFont;
  i :integer;
begin
  if not HasDoc then
    raise EPdfInvalidOperation.Create('document is null.');
    
  // if specified font exists in fontlist, return it. otherwise, create the font.
  result := nil;
  for i := 0 to FFontList.Count - 1 do
  begin
    FFont := TPdfFont(FFontList.Items[i]);
    if FFont.Name = FontName then
    begin
      result := FFont;
      Break;
    end;
  end;
  if result = nil then
    result := CreateFont(FontName);
end;

function TPdfDoc.GetXObject(AName: string): TPdfXObject;
var
  FXObject: TPdfXObject;
  i :integer;
begin
  // return the XObject which name is muched with specified name.
  result := nil;
  for i := 0 to FXObjectList.ItemCount - 1 do
  begin
    FXObject := TPdfXObject(FXObjectList.Items[i]);
    if TPdfName(FXObject.Attributes.ValueByName('Name')).Value = AName then
    begin
      result := FXObject;
      Break;
    end;
  end;
end;

function TPdfDoc.AddOutlineEntry(AParent: TPdfOutlineEntry; ATitle: string;
  AXPos, AYPos: Single; AOpened: boolean): TPdfOutlineEntry;
var
  FParent: TPdfDictionary;
  FOpened: integer;
begin
  if not HasDoc then
    raise EPdfInvalidOperation.Create('document is null.');
  FParent := AParent;
  if not Root.UseOutlines then
    Root.UseOutlines := true;
  if FParent = nil then
    FParent := Root.Outlines;
  if AOpened then
    FOpened := PDF_ENTRY_OPENED
  else
    FOpened := PDF_ENTRY_CLOSED;
  result := _OutlineEntry_Create(FParent, ATitle, FCanvas.Page, AXPos,
    AYPos, FOpened);
end;

procedure TPdfDoc.NewDoc;
begin
  {*
   * create new document.
   *}
  FreeDoc;
  FHeader := TPdfHeader.Create;
  FTrailer := TPdfTrailer.Create(FXref);
  FXref := TPdfXref.Create;
  FFontList := TList.Create;
  FXObjectList := TPdfArray.CreateArray(FXref);

  FRoot := TPdfCatalog.Create;
  FRoot.SetData(CreateCatalog);
  FRoot.UseOutlines := true;

  CreateInfo;
  FInfo.CreationDate := now;

  FCurrentPages := CreatePages(nil);
  FRoot.SetPages(FCurrentPages);

  FHasDoc := true;
end;

{$IFNDEF NOIMAGE}
procedure TPdfDoc.AddImage(AName: string; AImage: TGraphic; ImageClassName: string);
var
  FXObject: TPdfXObject;
  PdfImageCreator: TPdfImageCreator;
begin
  if GetXObject(AName) <> nil then
    raise Exception.CreateFmt('the image named %s is already exists..', [AName]);

  // create new image and regist to XObject table and Xref.
  PdfImageCreator := TPdfImageCreator(FindClass(ImageClassName).Create);
  try
    if PdfImageCreator = nil then
      raise Exception.Create('InvalidImageClassName:' + ImageClassName);
    FXObject := PdfImageCreator.CreateImage(FXref, AImage);
    FXref.AddObject(FXObject);
    RegisterXObject(FXObject, AName);
  finally
    PdfImageCreator.Free;
  end;
end;
{$ENDIF}

procedure TPdfDoc.AddPage;
var
  FPage: TPdfDictionary;
  FMediaBox: TPdfArray;
  FContents: TPdfStream;
  FResources: TPdfDictionary;
  FProcSet: TPdfArray;
  FFontArray: TPdfDictionary;
  FXObjectArray: TPdfDictionary;
  {$IFNDEF NOZLIB}
  FFilter: TPdfArray;
  {$ENDIF}
begin
  if FCurrentPages = nil then
    raise EPdfInvalidOperation.Create('internal error. current pages null.');

  // create new page object and add it to the current pages object.
  FPage := TPdfDictionary.CreateDictionary(FXref);
  FXref.AddObject(FPage);

  _Pages_AddKids(FCurrentPages, FPage);

  FPage.AddItem('Type', TPdfName.CreateName('Page'));
  FPage.AddItem('Parent', FCurrentPages);

  FMediaBox := TPdfArray.CreateArray(FXref);
  with FMediabox do
  begin
    AddItem(TPdfNumber.CreateNumber(0));
    AddItem(TPdfNumber.CreateNumber(0));
    AddItem(TPdfNumber.CreateNumber(DefaultPageWidth));
    AddItem(TPdfNumber.CreateNumber(DefaultPageHeight));
  end;
  FPage.AddItem('MediaBox', FMediaBox);

  FResources := TPdfDictionary.CreateDictionary(FXref);
  FPage.AddItem('Resources', FResources);

  FFontArray := TPdfDictionary.CreateDictionary(FXref);
  FResources.AddItem('Font', FFontArray);

  FXObjectArray := TPdfDictionary.CreateDictionary(FXref);
  FResources.AddItem('XObject', FXObjectArray);

  FProcSet := TPdfArray.CreateArray(FXref);
  with FProcSet do
  begin
    AddItem(TPdfName.CreateName('PDF'));
    AddItem(TPdfName.CreateName('Text'));
    AddItem(TPdfName.CreateName('ImageC'));
  end;
  FResources.AddItem('ProcSet', FProcSet);

  FContents := TPdfStream.CreateStream(FXref);
  FXref.AddObject(FContents);
  {$IFNDEF NOZLIB}
  FFilter := FContents.Attributes.PdfArrayByName('Filter');
  if FCompressionMethod = cmFlateDecode then
    FFilter.AddItem(TPdfName.CreateName('FlateDecode'));
  {$ENDIF}
  FPage.AddItem('Contents', FContents);

  FCanvas.SetPage(FPage);
end;

procedure TPdfDoc.FreeDoc;
var
  i: integer;
begin
  if FHasDoc then
  begin
    FXObjectList.Free;
    for i := FFontList.Count - 1 downto 0 do
      TObject(FFontList.Items[i]).Free;
    FFontList.Free;
    FXref.Free;
    FHeader.Free;
    FTrailer.Free;
    FHasDoc := false;
    if FInfo <> nil then
      FInfo.Free;
    FInfo := nil;
    FRoot.Free;
  end;
end;

procedure TPdfDoc.SaveToStream(AStream: TStream);
var
  i: integer;
  Pos: integer;
  PdfNumber: TPdfNumber;
begin
  if not HasDoc then
    raise EPdfInvalidOperation.Create('there is no document to save.');
  // write all objects to specified stream.
  FInfo.ModDate := Now;

  AStream.Position := 0;
  FHeader.WriteToStream(AStream);
  for i := 1 to FXref.ItemCount - 1 do
  begin
    Pos := AStream.Position;
    FXref.Items[i].Value.WriteValueToStream(AStream);
    FXref.Items[i].ByteOffset := Pos;
  end;
  FTrailer.XrefAddress := AStream.Position;
  FXref.WriteToStream(AStream);
  PdfNumber := FTrailer.Attributes.PdfNumberByName('Size');
  PdfNumber.Value := FXref.ItemCount;
  FTrailer.WriteToStream(AStream);
end;

{ TPdfCanvasAttribute }
procedure TPdfCanvasAttribute.SetFont(Value: TPdfFont);
begin
  if FFont <> Value then
    FFont := Value;
end;

procedure TPdfCanvasAttribute.SetWordSpace(Value: Single);
begin
  if Value < 0 then
    raise EPdfInvalidValue.Create('invalid word space');
  if Value <> FWordSpace then
    FWordSpace := Value;
end;

procedure TPdfCanvasAttribute.SetCharSpace(Value: Single);
begin
  if Value <> FCharSpace then
    FCharSpace := Value;
end;

procedure TPdfCanvasAttribute.SetFontSize(Value: Single);
begin
  if Value < 0 then
    raise EPdfInvalidValue.Create('invalid font size');
  if Value <> FFontSize then
    FFontSize := Value;
end;

procedure TPdfCanvasAttribute.SetHorizontalScaling(Value: Word);
begin
  if Value > PDF_MAX_HORIZONTALSCALING then
    raise EPdfInvalidValue.Create('invalid font size');
  if Value <> FHorizontalScaling then
    FHorizontalScaling := Value;
end;

function TPdfCanvasAttribute.TextWidth(Text: string): Single;
var
  i: integer;
  ch: char;
  tmpWidth: Single;
begin
  result := 0;

  // calculate width of specified text from current attributes
  for i := 1 to Length(Text) do
  begin
    ch := Text[i];
    tmpWidth := FFont.GetCharWidth(Text, i) * FFontSize / 1000;
    if FHorizontalScaling <> 100 then
      tmpWidth := tmpWidth * FHorizontalScaling / 100;
    if tmpWidth > 0 then
      tmpWidth := tmpWidth + FCharSpace
    else
      tmpWidth := 0;
    if (ch = ' ') and (FWordSpace > 0) and (i <> Length(Text)) then
      tmpWidth := tmpWidth + FWordSpace;

    result := result + tmpWidth;
  end;
  result := result - FCharSpace;
end;

function TPdfCanvasAttribute.MesureText(Text: string; Width: Single): integer;
var
  i: integer;
  ch: char;
  tmpWidth: Single;
  tmpTotalWidth: Single;
begin
  result := 0;
  tmpTotalWidth := 0;

  // calculate number of charactor contain in thw specified width.
  for i := 1 to Length(Text) do
  begin
    ch := Text[i];
    tmpWidth := FFont.GetCharWidth(Text, i) * FFontSize / 1000;
    if FHorizontalScaling <> 100 then
      tmpWidth := tmpWidth * FHorizontalScaling / 100;
    if tmpWidth > 0 then
      tmpWidth := tmpWidth + FCharSpace
    else
      tmpWidth := 0;
    if (ch = ' ') and (FWordSpace > 0) and (i <> Length(Text)) then
      tmpWidth := tmpWidth + FWordSpace;

    tmpTotalWidth := tmpTotalWidth + tmpWidth;
    if tmpTotalWidth > Width then
      Break;
    inc(result);
  end;
end;

{ TPdfCanvas }

constructor TPdfCanvas.Create(APdfDoc: TPdfDoc);
begin
  FPdfDoc := APdfDoc;
  FPage := nil;
  FContents := nil;
  FAttr := TPdfCanvasAttribute.Create;
end;

destructor TPdfCanvas.Destroy;
begin
  FAttr.Free;
  inherited;
end;

procedure TPdfCanvas.SetPageWidth(AValue: integer);
var
  FMediaBox: TPdfArray;
begin
  FMediaBox := TPdfArray(FPage.ValueByName('MediaBox'));
  if FMediaBox <> nil then
    TPdfNumber(FMediaBox.Items[2]).Value := AValue
  else
    EPdfInvalidOperation.Create('Can not chenge width of this page..');
end;

procedure TPdfCanvas.SetPageHeight(AValue: integer);
var
  FMediaBox: TPdfArray;
begin
  FMediaBox := TPdfArray(FPage.ValueByName('MediaBox'));
  if FMediaBox <> nil then
    TPdfNumber(FMediaBox.Items[3]).Value := AValue
  else
    EPdfInvalidOperation.Create('Can not chenge width of this page..');
end;

function TPdfCanvas.GetPageWidth: Integer;
var
  FMediaBox: TPdfArray;
begin
  FMediaBox := TPdfArray(FPage.ValueByName('MediaBox'));
  if FMediaBox <> nil then
    result := TPdfNumber(FMediaBox.Items[2]).Value
  else
    result := FPdfDoc.DefaultPageWidth;
end;

function TPdfCanvas.GetPageHeight: Integer;
var
  FMediaBox: TPdfArray;
begin
  FMediaBox := TPdfArray(FPage.ValueByName('MediaBox'));
  if FMediaBox <> nil then
    result := TPdfNumber(FMediaBox.Items[3]).Value
  else
    result := FPdfDoc.DefaultPageHeight;
end;

function TPDFCanvas.GetColorStr(Color: TColor): string;
var
  X: array[0..3] of Byte;
  i: integer;
begin
  i := ColorToRGB(Color);
  Move(i, x[0], 4);
  result := FloatToStrR(X[0] / 255) + ' ' +
            FloatToStrR(X[1] / 255) + ' ' +
            FloatToStrR(X[2] / 255);
end;

procedure TPdfCanvas.SetPage(APage: TPdfDictionary);
  procedure GetCurrentFont;
  var
    AFont: TPdfName;
  begin
    AFont := FPage.PdfNameByName('_Font');
    if AFont <> nil then
    begin
      FAttr.Font := FPdfDoc.GetFont(AFont.Value);
      FAttr.FontSize := FPage.PdfNumberByName('_Font_Size').Value;
      FAttr.WordSpace := FPage.PdfRealByName('_Word_Space').Value;
      FAttr.CharSpace := FPage.PdfRealByName('_Char_Space').Value;
      FAttr.HorizontalScaling := FPage.PdfNumberByName('_HScalling').Value;
    end
    else
    begin
      SetFont(PDF_DEFAULT_FONT, PDF_DEFAULT_FONT_SIZE);
      FAttr.CharSpace := 0;
      FAttr.WordSpace := 0;
      FAttr.HorizontalScaling := 100;
    end;
  end;
begin
  // save current canvas attributes to internal objects.
  if FPage <> nil then
  begin
    FPage.AddInternalItem('_Font', TPdfName.CreateName(FAttr.Font.Name));
    FPage.AddInternalItem('_Font_Size', TPdfReal.CreateReal(FAttr.FontSize));
    FPage.AddInternalItem('_Word_Space', TPdfReal.CreateReal(FAttr.WordSpace));
    FPage.AddInternalItem('_Char_Space', TPdfReal.CreateReal(FAttr.CharSpace));
    FPage.AddInternalItem('_HScalling', TPdfNumber.CreateNumber(FAttr.HorizontalScaling));
  end;
  FPage := APage;
  FContents := TPdfStream(FPage.ValueByName('Contents'));
  GetCurrentFont;
end;

procedure TPdfCanvas.SetFont(AName: string; ASize: Single);
var
  FFont: TPdfFont;
  FFontList: TPdfDictionary;
  FFontName: string;
begin
  // get font object from pdfdoc object, then find fontlist from page object
  // by internal name. if font is not registered, register it to page object.
  FFont := FPdfDoc.GetFont(AName);
  if (FAttr.Font = FFont) and (FAttr.FontSize = ASize) then Exit;
  FFontList := _Page_GetResources(FPage, 'Font');
  FFontName := FFont.Data.PdfNameByName('Name').Value;
  if FFontList.ValueByName(FFontName) = nil then
    FFontList.AddItem(FFontName, FFont.Data);
  if FContents <> nil then
    SetFontAndSize('/' + FFontName, ASize);
  FAttr.Font := FFont;
  FAttr.FontSize := ASize;
end;

procedure TPdfCanvas.DrawXObject(X, Y, AWidth, AHeight: Single;
   AXObjectName: string);
var
  XObject: TPdfXObject;
  FXObjectList: TPdfDictionary;
begin
  // drawing object must be registered. check object name.
  XObject := FPdfDoc.GetXObject(AXObjectName);
  if XObject = nil then
    raise EPdfInvalidValue.CreateFmt('XObject not found: %s', [AXObjectName]);

  FXObjectList := _Page_GetResources(FPage, 'XObject');
  if FXObjectList.ValueByName(AXObjectName) = nil then
    FXObjectList.AddItem(AXObjectName, XObject);

  GSave;
  Concat(AWidth, 0, 0, AHeight, X, Y);
  ExecuteXObject(XObject.Attributes.PdfNameByName('Name').Value);
  GRestore;
end;

  {* Special Graphics State *}

procedure TPdfCanvas.GSave;
begin
  _WriteString('q'#10, FContents.Stream);
end;

procedure TPdfCanvas.GRestore;
begin
  _WriteString('Q'#10, FContents.Stream);
end;

procedure TPdfCanvas.Concat(a, b, c, d, e, f: Single);
var
  S: string;
begin
  S := FloatToStr(a) + ' ' +
       FloatToStr(b) + ' ' +
       FloatToStr(c) + ' ' +
       FloatToStr(d) + ' ' +
       FloatToStr(e) + ' ' +
       FloatToStr(f) + ' cm'#10;
  _WriteString(S, FContents.Stream);
end;

  {* General Graphics State *}
procedure TPdfCanvas.SetFlat(flatness: Byte);
var
  S: string;
begin
  S := IntToStr(flatness) + ' i'#10;
  _WriteString(S, FContents.Stream);
end;

procedure TPdfCanvas.SetLineCap(linecap: TLineCapStyle);
var
  S: string;
begin
  S := IntToStr(ord(linecap)) + ' J'#10;
  _WriteString(S, FContents.Stream);
end;

procedure TPdfCanvas.SetDash(aarray: array of Byte; phase: Byte);
var
  S: string;
  i: integer;
begin
  S := '[';
  if (High(aarray) >= 0) and (aarray[0] <> 0) then   // compatibility fo delphi 3
    for i := 0 to High(aarray) do
      S := S + IntToStr(aarray[i]) + ' ';
  S := S + '] ' + IntToStr(phase) + ' d'#10;
  _WriteString(S, FContents.Stream);
end;

procedure TPdfCanvas.SetLineJoin(linejoin: TLineJoinStyle);
var
  S: string;
begin
  S := IntToStr(ord(linejoin)) + ' j'#10;
  _WriteString(S, FContents.Stream);
end;

procedure TPdfCanvas.SetLineWidth(linewidth: Single);
var
  S: string;
begin
  S := FloatToStrR(linewidth) + ' w'#10;
  _WriteString(S, FContents.Stream);
end;

procedure TPdfCanvas.SetMiterLimit(miterlimit: Byte);
var
  S: string;
begin
  S := IntToStr(miterlimit) + ' M'#10;
  _WriteString(S, FContents.Stream);
end;

  {* Paths *}
procedure TPdfCanvas.MoveTo(x, y: Word);
var
  S: string;
begin
  S := IntToStr(x) + ' ' + IntToStr(y) + ' m'#10;
  _WriteString(S, FContents.Stream);
end;

procedure TPdfCanvas.LineTo(x, y: Word);
var
  S: string;
begin
  S := IntToStr(x) + ' ' + IntToStr(y) + ' l'#10;
  _WriteString(S, FContents.Stream);
end;

procedure TPdfCanvas.CurveToC(x1, y1, x2, y2, x3, y3: Word);
var
  S: string;
begin
  S := IntToStr(x1) + ' ' +
       IntToStr(y1) + ' ' +
       IntToStr(x2) + ' ' +
       IntToStr(y2) + ' ' +
       IntToStr(x3) + ' ' +
       IntToStr(y3) + ' c'#10;
  _WriteString(S, FContents.Stream);
end;

procedure TPdfCanvas.CurveToV(x2, y2, x3, y3: Word);
var
  S: string;
begin
  S := IntToStr(x2) + ' ' +
       IntToStr(y2) + ' ' +
       IntToStr(x3) + ' ' +
       IntToStr(y3) + ' v'#10;
  _WriteString(S, FContents.Stream);
end;

procedure TPdfCanvas.CurveToY(x1, y1, x3, y3: Word);
var
  S: string;
begin
  S := IntToStr(x1) + ' ' +
       IntToStr(y1) + ' ' +
       IntToStr(x3) + ' ' +
       IntToStr(y3) + ' y'#10;
  _WriteString(S, FContents.Stream);
end;

procedure TPdfCanvas.Rectangle(x, y, width, height: Word);
var
  S: string;
begin
  S := IntToStr(x) + ' ' +
       IntToStr(y) + ' ' +
       IntToStr(width) + ' ' +
       IntToStr(height) + ' re'#10;
  _WriteString(S, FContents.Stream);
end;

procedure TPdfCanvas.Closepath;
begin
  _WriteString('h'#10, FContents.Stream);
end;

procedure TPdfCanvas.NewPath;
begin
  _WriteString('n'#10, FContents.Stream);
end;

procedure TPdfCanvas.Stroke;
begin
  _WriteString('S'#10, FContents.Stream);
end;

procedure TPdfCanvas.ClosePathStroke;
begin
  _WriteString('s'#10, FContents.Stream);
end;

procedure TPdfCanvas.Fill;
begin
  _WriteString('f'#10, FContents.Stream);
end;

procedure TPdfCanvas.Eofill;
begin
  _WriteString('f*'#10, FContents.Stream);
end;

procedure TPdfCanvas.FillStroke;
begin
  _WriteString('B'#10, FContents.Stream);
end;

procedure TPdfCanvas.ClosepathFillStroke;
begin
  _WriteString('b'#10, FContents.Stream);
end;

procedure TPdfCanvas.EofillStroke;
begin
  _WriteString('B*'#10, FContents.Stream);
end;

procedure TPdfCanvas.ClosepathEofillStroke;
begin
  _WriteString('b*'#10, FContents.Stream);
end;

procedure TPdfCanvas.Clip;
begin
  _WriteString('W'#10, FContents.Stream);
end;

procedure TPdfCanvas.Eoclip;
begin
  _WriteString('W*'#10, FContents.Stream);
end;

  {* Test state *}
procedure TPdfCanvas.SetCharSpace(charSpace: Single);
begin
  if FAttr.CharSpace = charSpace then Exit;
  FAttr.SetCharSpace(charSpace);
  if Contents <> nil then
    _WriteString(FloatToStr(charSpace) + ' Tc'#10, FContents.Stream);
end;

procedure TPdfCanvas.SetWordSpace(wordSpace: Single);
begin
  if FAttr.WordSpace = wordSpace then Exit;
  FAttr.SetWordSpace(wordSpace);
  if Contents <> nil then
    _WriteString(FloatToStr(wordSpace) + ' Tw'#10, FContents.Stream);
end;

procedure TPdfCanvas.SetHorizontalScaling(hScaling: Word);
begin
  if FAttr.HorizontalScaling = hScaling then Exit;
  FAttr.SetHorizontalScaling(hScaling);
  _WriteString(IntToStr(hScaling) + ' Tz'#10, FContents.Stream);
end;

procedure TPdfCanvas.SetLeading(leading: Single);
begin
  _WriteString(FloatToStr(leading) + ' TL'#10, FContents.Stream);
end;

procedure TPdfCanvas.SetFontAndSize(fontname: string; size: Single);
var
  S: string;
begin
  S := fontname + ' ' +
       FloatToStr(size) + ' Tf'#10;
  _WriteString(S, FContents.Stream);
end;

procedure TPdfCanvas.SetTextRenderingMode(mode: TTextRenderingMode);
begin
  _WriteString(IntToStr(ord(mode)) + ' Tr'#10, FContents.Stream);
end;

procedure TPdfCanvas.SetTextRise(rise: Word);
begin
  _WriteString(IntToStr(rise) + ' Ts'#10, FContents.Stream);
end;

procedure TPdfCanvas.BeginText;
begin
  _WriteString('BT'#10, FContents.Stream);
end;

procedure TPdfCanvas.EndText;
begin
  _WriteString('ET'#10, FContents.Stream);
end;

procedure TPdfCanvas.MoveTextPoint(tx, ty: Word);
var
  S: string;
begin
  S := IntToStr(tx) + ' ' +
       IntToStr(ty) + ' Td'#10;
  _WriteString(S, FContents.Stream);
end;

procedure TPdfCanvas.SetTextMatrix(a, b, c, d, x, y: Word);
var
  S: string;
begin
  S := IntToStr(a) + ' ' +
       IntToStr(b) + ' ' +
       IntToStr(c) + ' ' +
       IntToStr(d) + ' ' +
       IntToStr(x) + ' ' +
       IntToStr(y) + ' Tm'#10;
  _WriteString(S, FContents.Stream);
end;

procedure TPdfCanvas.MoveToNextLine;
begin
  _WriteString('T*'#10, FContents.Stream);
end;

procedure TPdfCanvas.ShowText(s: string);
var
  FString: string;
begin
  if _HasMultiByteString(s) then
    FString := '<' + _StrToHex(s) + '>'
  else
    FString := '(' + _EscapeText(s) + ')';
  _WriteString(FString + ' Tj'#10, FContents.Stream);
end;

{ TPdfCanvas common routine }
function TPdfCanvas.TextWidth(Text: string): Single;
begin
  result := FAttr.TextWidth(Text);
end;

function TPdfCanvas.MesureText(Text: string; AWidth: Single): integer;
begin
  result := FAttr.MesureText(Text, AWidth);
end;

procedure TPdfCanvas.ShowTextNextLine(s: string);
var
  FString: string;
begin
  if _HasMultiByteString(s) then
    FString := '<' + _StrToHex(s) + '>'
  else
    FString := '(' + _EscapeText(s) + ')';
  _WriteString(FString + ' '''#10, FContents.Stream);
end;

procedure TPdfCanvas.ExecuteXObject(xObject: string);
var
  S: string;
begin
  S := '/' + xObject + ' Do'#10;
  _WriteString(S, FContents.Stream);
end;

procedure TPdfCanvas.SetRGBFillColor(Value: TColor);
var
  S: string;
begin
  S := GetColorStr(Value) + ' rg'#10;
  _WriteString(S, FContents.Stream);
end;

procedure TPdfCanvas.SetRGBStrokeColor(Value: TColor);
var
  S: string;
begin
  S := GetColorStr(Value) + ' RG'#10;
  _WriteString(S, FContents.Stream);
end;

function TPdfCanvas.GetDoc: TPdfDoc;
begin
  result := FPdfDoc;
end;

{ TPdfDictionaryWrapper }
procedure TPdfDictionaryWrapper.SetData(AData: TPdfDictionary);
begin
  FData := AData;
end;

function TPdfDictionaryWrapper.GetHasData: boolean;
begin
  result := (FData = nil);
end;

{ TPdfInfo }
procedure TPdfInfo.SetAuthor(Value: string);
begin
  FData.AddItem('Author', TPdfText.CreateText(Value));
end;

procedure TPdfInfo.SetCreationDate(Value: TDateTime);
begin
  FData.AddItem('CreationDate', TPdfText.CreateText(_DateTimeToPdfDate(Value)));
end;

procedure TPdfInfo.SetModDate(Value: TDateTime);
begin
  FData.AddItem('ModDate', TPdfText.CreateText(_DateTimeToPdfDate(Value)));
end;

procedure TPdfInfo.SetCreator(Value: string);
begin
  FData.AddItem('Creator', TPdfText.CreateText(Value));
end;

procedure TPdfInfo.SetTitle(Value: string);
begin
  FData.AddItem('Title', TPdfText.CreateText(Value));
end;

procedure TPdfInfo.SetSubject(Value: string);
begin
  FData.AddItem('Subject', TPdfText.CreateText(Value));
end;

procedure TPdfInfo.SetKeywords(Value: string);
begin
  FData.AddItem('Keywords', TPdfText.CreateText(Value));
end;

function TPdfInfo.GetAuthor: string;
begin
  if FData.ValueByName('Author') <> nil then
    result := FData.PdfTextByName('Author').Value
  else
    result := '';
end;

function TPdfInfo.GetCreationDate: TDateTime;
begin
  if FData.ValueByName('CreationDate') <> nil then
  try
    result := _PdfDateToDateTime(FData.PdfTextByName('CreationDate').Value);
  except
    result := 0;
  end
  else
    result := 0;
end;

function TPdfInfo.GetModDate: TDateTime;
begin
  if FData.ValueByName('ModDate') <> nil then
  try
    result := _PdfDateToDateTime(FData.PdfTextByName('ModDate').Value);
  except
    result := 0;
  end
  else
    result := 0;
end;

function TPdfInfo.GetCreator: string;
begin
  if FData.ValueByName('Creator') <> nil then
    result := FData.PdfTextByName('Creator').Value
  else
    result := '';
end;

function TPdfInfo.GetTitle: string;
begin
  if FData.ValueByName('Title') <> nil then
    result := FData.PdfTextByName('Title').Value
  else
    result := '';
end;

function TPdfInfo.GetSubject: string;
begin
  if FData.ValueByName('Subject') <> nil then
    result := FData.PdfTextByName('Subject').Value
  else
    result := '';
end;

function TPdfInfo.GetKeywords: string;
begin
  if FData.ValueByName('Keywords') <> nil then
    result := FData.PdfTextByName('Keywords').Value
  else
    result := '';
end;

{ TPdfCatalog }

procedure TPdfCatalog.SetPageMode(Value: TPdfPageMode);
var
  FPageMode: TPdfName;
begin
  FPageMode := TPdfName(FData.ValueByName('PageMode'));
  if (FPageMode = nil) or (not (FPageMode is TPdfName)) then
    FData.AddItem('PageMode', TPdfName.CreateName(PDF_PAGE_MODE_NAMES[Ord(Value)]))
  else
    FPageMode.Value := PDF_PAGE_MODE_NAMES[Ord(Value)];
end;

function TPdfCatalog.GetPageMode: TPdfPageMode;
var
  FPageMode: TPdfName;
  S: string;
  i: integer;
begin
  result := pmUseNone;
  FPageMode := TPdfName(FData.ValueByName('PageMode'));
  if (FPageMode = nil) or (not (FPageMode is TPdfName)) then
    Exit
  else
  begin
    S := FPageMode.Value;
    for i := 0 to High(PDF_PAGE_MODE_NAMES) do
      if PDF_PAGE_MODE_NAMES[i] = S then
      begin
        result := TPdfPageMode(i);
        Break;
      end;
  end;
end;

procedure TPdfCatalog.SetUseOutlines(Value: boolean);
begin
  if Value = UseOutlines then
    Exit;
  if Value then
    CreateOutlines
  else
  begin
    FData.RemoveItem('Outlines');
    FOutlines := nil;
  end;
end;

function TPdfCatalog.GetUseOutlines: boolean;
begin
  if FData.ValueByName('Outlines') <> nil then
    result := true
  else
    result := false;
end;

function TPdfCatalog.GetPages: TPdfDictionary;
begin
  result := TPdfDictionary(FData.ValueByName('Pages'));
  if result = nil then
    raise EPdfInvalidOperation.Create('ERROR: page object is null..');
end;

procedure TPdfCatalog.SetPages(APage: TPdfDictionary);
begin
  if _GetTypeOf(APage) = 'Pages' then
  begin
    FData.AddItem('Pages', APage);
    FPages := APage;
  end;
end;

function TPdfCatalog.GetOutlines: TPdfDictionary;
begin
  if FOutlines = nil then
    FOutlines := TPdfDictionary(FData.ValueByName('Outlines'));
  if FOutlines = nil then
    raise EPdfInvalidOperation.Create('This Document has no outlines..');
  result := FOutlines;
end;

procedure TPdfCatalog.CreateOutlines;
var
  FOutlines: TPdfOutlines;
begin
  FOutlines := TPdfOutlines.CreateDictionary(FData.ObjectMgr);
  FData.ObjectMgr.AddObject(FOutlines);
  with FOutLines do
  begin
    AddItem('Type', TPdfName.CreateName('Outlines'));
    AddItem('Count', TPdfNumber.CreateNumber(0));
    AddInternalItem('Opened', TPdfNumber.CreateNumber(PDF_ENTRY_OPENED));
  end;
  FData.AddItem('Outlines', FOutlines);
end;

{ TPdfFont }
procedure TPdfFont.AddStrElements(ADic: TPdfDictionary;
  ATable: array of TPDF_STR_TBL);
var
  i: integer;
begin
  { utility routine for making font dictinary. }
  for i := 0 to High(ATable) do
    ADic.AddItem(ATable[i].KEY, TPdfName.CreateName(ATable[i].VAL));
end;

procedure TPdfFont.AddIntElements(ADic: TPdfDictionary;
  ATable: array of TPDF_INT_TBL);
var
  i: integer;
begin
  { utility routine for making font dictinary. }
  for i := 0 to High(ATable) do
    ADic.AddItem(ATable[i].KEY, TPdfNumber.CreateNumber(ATable[i].VAL));
end;

function TPdfFont.GetCharWidth(AText: string; APos: integer): integer;
begin
  result := 0;
end;

constructor TPdfFont.Create(AXref: TPdfXref; AName: string);
begin
  inherited Create;
  FName := AName;
end;

{$IFNDEF NOIMAGE}

{ TPdfImageCreator }
function TPdfImageCreator.CreateImage(AXref: TPdfXref; AImage: TGraphic): TPdfXObject;
begin
  result := nil;
end;

constructor TPdfImageCreator.Create;
begin
  inherited;
end;
{$ENDIF}

end.
